__all__ = [
    'Task',
    'ForkTask',
    'MapTask',
    'ReduceTask',
    'utils',
]

from sys import stderr

try:
    from tasker.tasks import torch
    __all__.extend('torch')
except RuntimeError:
    print(
        'Extra "pytorch" has not been initialized, please install torch and pytorch-ignite modules to enable this '
        'extra.',
        file=stderr
    )
from tasker.tasks import utils
from tasker.tasks.containers import Task, ForkTask, MapTask, ReduceTask



from collections import defaultdict
from logging import Logger
from typing import Text, List, Type
from inspect import signature, Parameter

from tasker.mixin import ProfileMixin, value
from tasker.storage import Storage
from tasker.typedef import Profile
from tasker.tasks import Task


def def_task(require: List[Text]=(), provide: List[Text]=(), remove: List[Text]=(), **kwargs):
    def name_mapping(name):
        return ''.join(map(lambda it: it.capitalize(), name.split('_'))) + 'Task'

    def wrapper(func):
        parameter_names = tuple(signature(func).parameters)
        assert 'ctx' in parameter_names or 'context' in parameter_names

        def new_invoke(self, profile: Profile, shared: Storage, logger: Logger):
            return func({
                'shared': shared,
                'logger': logger
            }, **profile)

        def new_require(self):
            return require

        def new_provide(self):
            return provide

        def new_remove(self):
            return remove

        def new_define(self):
            annotation_map = defaultdict(default_factory=lambda it: str, map={
                str: str,
                int: int,
                float: float,
                bool: bool,
                List[Text]: list,
                List[int]: list,
                List[float]: list,
                List[str]: list,
                List[bool]: list,
            })
            return tuple(map(
                lambda it: value(it.name, annotation_map[it.annotation]),
                filter(
                    lambda it: (it.kind == Parameter.POSITIONAL_OR_KEYWORD or it.kind == Parameter.KEYWORD_ONLY) and it.name not in ('ctx', 'context'),
                    signature(func).parameters.values()
                )
            ))

        task_cls = type(name_mapping(func.__name__), (Task, ProfileMixin), {
            'invoke': new_invoke,
            'require': new_require,
            'provide': new_provide,
            'remove': new_remove,
            'define': new_define,
        })
        return task_cls

    return wrapper





from tasker.storage import Storage
from uuid import uuid4
from datetime import datetime
from pathlib import Path
import pickle
from os import remove, listdir, makedirs


class HardPickleStorage(Storage):
    def __init__(self, **kwargs):
        super(HardPickleStorage, self).__init__(**kwargs)
        self.id = f'{datetime.now().strftime("%Y%m%d%H%M%S")}-{uuid4().hex}' if 'id' not in kwargs else kwargs['id']
        self.storage_folder = Path('.tasker') / 'storage' / 'pickle' / self.id
        try:
            makedirs(self.storage_folder)
        except Exception:
            pass

    def __getitem__(self, item):
        with open(self.storage_folder / item, 'rb') as fp:
            return pickle.load(fp)

    def __setitem__(self, key, value):
        with open(self.storage_folder / key, 'wb') as fp:
            return pickle.dump(value, fp)

    def __delitem__(self, key):
        return remove(self.storage_folder / key)

    def __contains__(self, item):
        return listdir(self.storage_folder).__contains__(item)

    def load(self):
        pass

    def dump(self):
        pass


class SoftPickleStorage(Storage):
    def __init__(self, **kwargs):
        super(SoftPickleStorage, self).__init__(**kwargs)
        self.id = f'{datetime.now().strftime("%Y%m%d%H%M%S")}-{uuid4().hex}' if 'id' not in kwargs else kwargs['id']
        self.storage_folder = Path('.tasker') / 'storage' / 'pickle' / self.id
        try:
            makedirs(self.storage_folder)
        except Exception:
            pass
        self.cache_dict = {}

    def __getitem__(self, item):
        return self.cache_dict.__getitem__(item)

    def __setitem__(self, key, value):
        return self.__setitem__(key, value)

    def __delitem__(self, key):
        return self.__delitem__(key)

    def __contains__(self, item):
        return self.__contains__(item)

    def load(self):
        for key in listdir(self.storage_folder):
            with open(self.storage_folder / key, 'rb') as fp:
                self.cache_dict[key] = pickle.load(fp)

    def dump(self):
        for key in self.cache_dict.keys():
            with open(self.storage_folder / key, 'wb') as fp:
                pickle.dump(self.cache_dict[key], fp)



from enum import Enum
from typing import Text, Type, List, Tuple
from os import path, makedirs
from logging import getLogger as get_logger


def import_reference(reference: Text) -> Type:
    path = reference.split('.')
    if reference and len(path) > 0:
        root = path[0]
        module = __import__(root)
        if len(path) == 1:
            return module
        else:
            for layer in path[1:]:
                module = getattr(module, layer, None)
                if module is None:
                    raise RuntimeError(f'Reference {reference} NOT found.')
            return module
    else:
        raise RuntimeError(f'Wrong reference {reference}.')


def extract_reference(reference: Text) -> Tuple[Text, List[Text]]:
    reference_splits = reference.split(':')
    return reference_splits[0], reference_splits[1:] if len(reference_splits) > 1 else []


class LocalPath(object):
    _logger = get_logger('tasker.utils.LocalPath')

    class CATEGORY(Enum):
        CACHE = 'cache'
        LOG = 'log'
        OUTPUT = 'output'
        STORAGE = 'storage'
        TEMP = 'temp'

    @classmethod
    def create(cls, category, *name: Text):
        assert isinstance(category, cls.CATEGORY)
        if not name:
            cls._logger.warning('You are operating in a root category folder! Please re-check it.')

        absolute_root = path.abspath(path.curdir)
        absolute_path = path.join(absolute_root, category.value, *name)

        if not path.exists(absolute_path):
            makedirs(absolute_path)

        return absolute_path
